!Do Metropolis-Hastings

module MonteCarlo
 use ParamDef
 use CalcLike
 use Random
 implicit none
 type(ParamSet)  Trial
 integer :: burn_in = 2 !Minimum to get sensible answers
 integer :: indep_sample = 0
  !number of iterations between dumping full model data. If zero then skip.
contains

 subroutine MCMCsample(Params, accept_num)
   real Like
   integer accept_num
   Type(ParamSet) Params, CurParams
   real CurLike, MaxLike
   logical accpt
   integer num_accept, num, mult, last_num
   double precision ranmar

   MaxLike = LogZero
   CurLike = StartLike
   CurParams = Params
   num_accept = 0
   last_num = 0
   num = 0 
   mult= 1

   do while (num_accept <= accept_num)

   num = num + 1
 
   if (indep_sample /= 0) then
      if (num_accept>= burn_in .and. mod(num, indep_sample)==0) then
       if (CurLike /= LogZero) call WriteIndepSample(CurParams, CurLike)
      end if
   end if

   call GetProposal(CurParams, Trial)
   
   Like = GetLogLike(Trial)

   if (Feedback > 1) write (*,*) 'Likelihood: ', Like, 'Current Like:', CurLike

   accpt = (Like /= logZero) .and. &
            (CurLike > Like .or. Like < CurLike + 20 .and. ranmar() < exp(min(0.,-Like+CurLike))) 
             !Include the min() so that compilers not doing optimal compilation doen't complain

   call AcceptReject(accpt, CurParams%Info, Trial%Info)

   if (accpt) then

      if (num_accept> burn_in) then
         call WriteParams(CurParams,real(mult),CurLike)
      end if
   
      CurLike = Like
      CurParams = Trial
      num_accept = num_accept + 1
      if (Like < MaxLike) MaxLike = Like
      mult=1


      if (Feedback > 1) write (*,*) num, ' accepting. ratio:', real(num_accept)/num

      if (logfile_unit /=0 .and. mod(num_accept,50) ==0) then
           write (logfile_unit,*) 'rat:',real(num_accept)/num, ' in ',num,', best: ',real(MaxLike)
           write (logfile_unit,*) 'local acceptance ratio:', 50./(num - last_num) 
           
           call FlushFile(logfile_unit)
           if (Feedback > 1) write (*,*) 'rat:',real(num_accept)/num, ' in ',num, &
                                                    ', best: ',real(MaxLike) 
           last_num = num
      end if
    else
       mult=mult+1
       if (Feedback > 1) write (*,*) num,' rejecting. ratio:', real(num_accept)/num
    end if
 

    if (CurLike /= logZero) then
      call AddMPIParams(CurParams%P,CurLike) 
      if (mod(num,50)==0) call CheckParamChange
    else
      if (num > 1000) stop 'MCMC.f90: Couldn''t start after 1000 tries - check starting ranges'
    end if

   end do

 end subroutine MCMCsample

 
end module MonteCarlo
