!Module for using 147k 2dF data 
!See http://www.roe.ac.uk/~wjp/2dFGRS/2dF_Pk.html - adapted for f90

module twodf
use settings
use cmbtypes
implicit none

 !Note all units are in k/h here

 logical :: use_2dF = .false.

 integer, parameter :: twodf_points = 32
 real :: twodf_P(twodf_points), twodf_k(twodf_points), twodf_Ninv(twodf_points,twodf_points)
 logical :: do_twodf_init  = .true.

contains

 subroutine twodf_init
   real sigma, model
   integer i
   
    if (Feedback > 0) write(*,*) 'reading: 2df data'
    call OpenTxtFile('data/twodf_june02.dat', tmp_file_unit)
    twodf_NInv = 0
    do i =1, twodf_points
      read (tmp_file_unit,*, end = 200, err=200) twodf_k(i), twodf_P(i), sigma, model 
    end do
    close(tmp_file_unit)
    call ReadMatrix('data/twodf_june02_matrix.dat',twodf_Ninv,twodf_points,twodf_points)

   do_twodf_init = .false.

    read (tmp_file_unit,*, end = 300, err=300) model 

200 stop 'Error reading 2dF file'
     
300 return

 end subroutine twodf_init

 function twodf_Wksq(k)
   real twodf_Wksq
   real, parameter :: kp1 = 1.17011e-05, kp2 = 9.33106e-09, norm = 1150292.2
   real k,k2

   k2 =k**2
   twodf_Wksq = norm/(1 + k2/kp1 + k2**2/kp2)

 end function twodf_Wksq

 function twodf_PconvW(Theory, rk)
   real twodf_PconvW 
   Type (CosmoTheory) Theory
   real rk
   
   real, parameter :: rkmax = 0.3
   integer, parameter :: nint2 = 18
   integer nint1
   real dmu,s, rmu
   real rk2,dk
   real xk,xk2,a, oa, sp, yk
   integer imu, ik


  
   dmu = 2./nint2
   rk2 = rk**2
   
   dk = 0.5*rkmax
   xk = 0; xk2 = 0
   s = 0
   !first and last points in trapezoidal rule
   do ik=1,2
      sp = 0
      do imu = 1, nint2
         rmu = -1 + (imu-0.5)*dmu
         yk = sqrt(rk2 + xk2 - 2*xk*rk*rmu)
         sp = sp + MatterPowerAt(Theory, yk)
      end do
      s = s + sp*xk2*twodf_Wksq(xk)
      xk = rkmax; xk2= xk**2
   end do
   
   a = s*dk
   
   !Loop over iterations till convergence

   nint1 = 1

   do 
      oa = a
      s = 0
      dk = rkmax / nint1
      do ik=1,nint1
         xk = (ik-0.5)*dk
         xk2=xk**2
         sp = 0
         do imu = 1, nint2
            rmu = -1 + (imu-0.5)*dmu
            yk = sqrt(rk2+xk2-2*xk*rk*rmu)
            sp = sp + MatterPowerAt(Theory, yk)
         end do
         s = s+ sp*xk2*twodf_Wksq(xk)
      end do
      a = (oa + s*dk)/2
      nint1 = nint1*2
      if (abs(oa-a) < 1e-4*a) exit
   end do

   twodf_PconvW = a*dmu

 end function twodf_PconvW

 function LSS_2dflike(CMB, Theory)
   Type (CMBParams) CMB
   Type (CosmoTheory) Theory
   real LSS_2dflike
   real th(twodf_points), sum1, chisq
   real Ninvth(twodf_points), Ninvd(twodf_points)
   integer i
   
    if (do_twodf_init) call twodf_init

    !$OMP PARALLEL DO DEFAULT(SHARED),SCHEDULE(STATIC), PRIVATE(i)
    do i=1, twodf_points
     th(i) = twodf_PconvW(Theory,twodf_k(i))
    end do
    !$OMP END PARALLEL DO

    !Analytic marginalization
 
    Ninvth = matmul(twodf_Ninv,th)
    Ninvd = matmul(twodf_Ninv,twodf_P)
  
    sum1 = sum(th*Ninvth) 
   ! CMB%norm(3) = sqrt(sum(th*Ninvd)/(sum1* 0.92**2) - 4./45*(CMB%omc+CMB%omb)**1.2) - &
   !                 (CMB%omc+CMB%omb)**0.6/3
   ! write (*,*) CMB%norm(3) 

    chisq = sum(twodf_P*Ninvd) - sum(th*Ninvd)**2/sum1 + log(sum1)
   

    LSS_2dflike = chisq/2

    if (Feedback>1) write(*,*) '2df chi-sq: ', chisq
   
 end function LSS_2dflike

 function LSSLnLike(CMB, Theory)
   Type (CMBParams) CMB
   Type (CosmoTheory) Theory
   real LSSLnLike

    if (use_2dF) then
     LSSLnLike = LSS_2dflike(CMB,Theory)
    else
     LSSLnLike = 0
    end if

 end function LSSLnLike



end module twodf
